#pragma once
#include <utility>

#include "latency_compensator.h"
#include "audio/misc/audio_buffer.h"
#include "audio/misc/circular_audio_buffer.h"
#include "misc/delegates.h"

class channel_interface;
class mixer_track;
class plugin_host;

enum class mixer_track_type {
    unknown,
    dummy,
    instrument,
};

struct mixer_track_link {
    mixer_track_link() = default;
    mixer_track_link(mixer_track* in_track, float in_send_level) : track(in_track), send_level(in_send_level) {
    }
    mixer_track* track = nullptr;
    float send_level = 1.0f;
};


class CORE_API mixer_track {
    friend class mixer;
public:
    explicit mixer_track(mixer_track_type in_type) : type_(in_type), id_(gen_uid()) {
        ui_buffers = std::make_shared<circular_buffer_vector_type>();
    }
    [[nodiscard]] uint64_t get_uid() const { return id_; }
    virtual ~mixer_track();

    void init();
    void add_effect(plugin_host* in_effect);
    void remove_effect(plugin_host* in_effect);
    void add_child(mixer_track* in_child, float in_send_level);
    void remove_child(mixer_track* in_child);

    void process(uint32_t in_frames);
    void post_process(uint32_t in_frames);

    virtual void rename(const std::string& in_name) = 0;
    [[nodiscard]] virtual std::string get_name() const = 0;

    [[nodiscard]] sample_t** get_headers() { return buffer_.get_headers(); }
    [[nodiscard]] const std::vector<sample_t*>& get_headers_vector() { return buffer_.get_headers_vector(); }
    void clear() { buffer_.clear(); }

    [[nodiscard]] mixer_track_type get_type() const { return type_; }

    [[nodiscard]] channel_interface* get_channel_interface() const { return channel_interface_; }

    void set_volume(float in_volume) { volume_ = in_volume; }
    [[nodiscard]] float get_volume() const { return volume_; }

    [[nodiscard]] int32_t get_latency() const { return compensator_.get_latency(); }
    [[nodiscard]] int32_t get_user_latency() const { return compensator_.get_user_latency(); }
    [[nodiscard]] int32_t get_real_latency() const { return compensator_.get_real_latency(); }
    void set_user_latency(int32_t in_latency) { compensator_.set_latency(in_latency); }

    [[nodiscard]] std::string get_imgui_id() const {
        return get_name() + "##" + std::to_string(id_);
    }

    std::vector<plugin_host*> effects{};
    std::vector<mixer_track_link> children{};

    // UI
    std::shared_ptr<circular_buffer_vector_type> ui_buffers;
protected:
    virtual void on_latency_changed(int32_t in_latency);
    latency_compensator compensator_;
private:
    audio_buffer buffer_;
    audio_buffer temp_mix_buffer_;
    const mixer_track_type type_;
    float volume_ = 1.0f;
    channel_interface* channel_interface_ = nullptr;
    const uint64_t id_ = 0;
};

class instrument_track : public mixer_track {
public:
    static constexpr auto type = mixer_track_type::instrument;
    explicit instrument_track(plugin_host* in_instrument);

    void rename(const std::string& in_name) override;
    [[nodiscard]] std::string get_name() const override;

    [[nodiscard]] plugin_host* get_instrument() const { return instrument_; }
private:
    void on_latency_changed(int32_t in_latency) override;
    plugin_host* instrument_;
};

class dummy_track : public mixer_track {
public:
    static constexpr auto type = mixer_track_type::dummy;
    dummy_track() : mixer_track(mixer_track_type::dummy) {}

    void rename(const std::string& in_name) override { name_ = in_name; }
    [[nodiscard]] std::string get_name() const override { return name_; }
private:
    std::string name_;
};

CORE_API void delete_effect(mixer_track* track, plugin_host* host);
